/*
 * The Gold Partition Conjecture for K-thin posets
 *
 * isom.c
 * Graph isomorphism routine - Ullman's algorithm
 */

/* Copyright 2005, Marcin Peczarski. All Rights Reserved.
 *
 * This program is free software. You may use this software, and copy,
 * redistribute, and/or modify it for non profit works.
 * Please do not remove this copyright notice.
 */

#include "gpc.h"

#define ra_1 g1->r[0]
#define ra_2 g2->r[0]

BOOL isomorphic(const POSET *g1, const POSET *g2) {
  int  c10[N], c11[N], c20[N], c21[N];
  int  i, j, m, n, k, l[N]; /* k is perm_len, l[k] is node2 */
  BOOL possible;
  SET  MT[N][N], j_set;

  n = g1->n;
  m = g1->m;
  #ifdef _CHECK_CORRECTNESS_
  if (n != g2->n || m != g2->m)
    fatal_error("isomorphic, n1=%d m1=%d n2=%d m2=%d", n, m, g2->n, g2->m);
  #endif
  for (i = 0; i < n; i++) {
    c10[i] = cardinality(g1->r[0][i]);
    c20[i] = cardinality(g2->r[0][i]);
    c11[i] = cardinality(g1->r[1][i]);
    c21[i] = cardinality(g2->r[1][i]);
  }
  for (i = 0; i < m; i++) {
    MT[0][i] = EMPTY_SET;
    for (j = 0; j < m; j++)
      if (c10[i] == c20[j] && c11[i] == c21[j])
        MT[0][i] |= pof2(j);
    if (MT[0][i] == EMPTY_SET)
      return FALSE;
  }
  for (i = m; i < n; i++) {
    MT[0][i] = EMPTY_SET;
    for (j = m; j < n; j++)
      if (c10[i] == c20[j] && c11[i] == c21[j])
        MT[0][i] |= pof2(j);
    if (MT[0][i] == EMPTY_SET)
      return FALSE;
  }
  k = 0;
  for (i = 0; i < n; i++)
    l[i] = 0;
  for (;;) {
    possible = FALSE;
    if (in(MT[k][k], l[k])) {
      possible = TRUE;
      for (i = k + 1; i < n; i++) {
        j_set = MT[k + 1][i] = MT[k][i]; /* & ~pof2(l[k]); */
        /* If there exists only one of two edges:
           (i,k) in graph 1 or (j,l) in graph 2
           or exists only one of two edges:
           (k,i) in graph 1 or (l,j) in graph 2
           then isomorphism can't map vertex i to vertex j. */
        while (j_set) {
          j = min_element(j_set);
          if (in(ra_1[i], k) != in(ra_2[j], l[k]) ||
              in(ra_1[k], i) != in(ra_2[l[k]], j))
            MT[k + 1][i] ^= pof2(j);
          j_set ^= pof2(j);
        }
        MT[k + 1][i] &= ~pof2(l[k]);
        if (MT[k + 1][i] == EMPTY_SET) {
          possible = FALSE;
          break;
        }
      }
    }
    if (possible) {
      if (k == n - 2)
        return TRUE;
      else
        k++;
    }
    else {
      while (k >= 0 && l[k] >= n - 1) {
        l[k] = 0;
        k--;
      }
      if (k >= 0)
        l[k]++;
      else
        return FALSE;
    }
  }
}
